// Copyright 2022 Dennis Hezel
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef AGRPC_AGRPC_CANCEL_SAFE_HPP
#define AGRPC_AGRPC_CANCEL_SAFE_HPP

#include <agrpc/detail/asio_forward.hpp>
#include <agrpc/detail/config.hpp>

#ifdef AGRPC_ASIO_HAS_CANCELLATION_SLOT

#include <agrpc/default_completion_token.hpp>
#include <agrpc/detail/cancel_safe.hpp>
#include <agrpc/detail/tuple.hpp>
#include <agrpc/detail/type_erased_completion_handler.hpp>
#include <agrpc/detail/utility.hpp>
#include <agrpc/detail/work_tracking_completion_handler.hpp>

#include <cassert>
#include <optional>

AGRPC_NAMESPACE_BEGIN()

template <class CompletionSignature>
class CancelSafe;

/**
 * @brief Cancellation safety for asynchronous operations
 *
 * This class provides a completion token that can be used to initiate asynchronous operations in a cancellation safe
 * manner. A second method of this class is then used to wait for the operation to complete. Cancelling the wait
 * will not cancel the underlying operation but still invoke the completion handler with
 * `asio::error::operation_aborted`. This can be useful in combination with `asio::parallel_group` or
 * `asio::awaitable_operators`, e.g. to perform an action every 100ms while waiting for a server-stream:
 *
 * @snippet client.cpp cancel-safe-server-streaming
 *
 * @tparam CompletionArgs The arguments of the completion signature. E.g. for `asio::steady_timer::async_wait` the
 * completion arguments would be `boost::system::error_code`.
 *
 * @since 1.6.0 (and Boost.Asio 1.77.0)
 */
template <class... CompletionArgs>
class CancelSafe<void(CompletionArgs...)>
{
  private:
    using CompletionSignature = detail::PrependErrorCodeToSignatureT<void(CompletionArgs...)>;
    using Result = detail::Tuple<CompletionArgs...>;

    struct Initiator;

  public:
    /**
     * @brief The type of completion token used to initiate asynchronous operations
     */
    class CompletionToken
    {
      public:
        void operator()(CompletionArgs... completion_args)
        {
            if (auto ch = self_.completion_handler_.release())
            {
                detail::complete_successfully(std::move(ch), static_cast<CompletionArgs&&>(completion_args)...);
            }
            else
            {
                self_.result_.emplace(Result{static_cast<CompletionArgs&&>(completion_args)...});
            }
        }

      private:
        friend agrpc::CancelSafe<void(CompletionArgs...)>;

        explicit CompletionToken(CancelSafe& self) noexcept : self_(self) {}

        CancelSafe& self_;
    };

    /**
     * @brief Create a completion token to initiate asynchronous operations
     *
     * Thread-safe
     */
    [[nodiscard]] CompletionToken token() noexcept { return CompletionToken{*this}; }

    [[deprecated(
        "This function has always reported whether a wait is currently pending and not whether the asynchronous "
        "operation initiated with the CompletionToken has completed. It has therefore been renamed to "
        "`is_wait_pending()`.")]] [[nodiscard]] bool
    is_running() const noexcept
    {
        return is_wait_pending();
    }

    /**
     * @brief Is a wait currently pending?
     *
     * Thread-safe
     */
    [[nodiscard]] bool is_wait_pending() const noexcept { return bool{completion_handler_}; }

    /**
     * @brief Wait for the asynchronous operation to complete
     *
     * Only one call to `wait()` may be outstanding at a time. Waiting for an already completed operation will
     * immediately invoke the completion handler in a manner equivalent to using `asio::post`.
     *
     * Thread-unsafe with regards to successful completion of the asynchronous operation.
     *
     * @param token Completion token that matches the completion args. Either `void(error_code, CompletionArgs...)` if
     * the first argument in CompletionArgs is not `error_code` or `void(CompletionArgs...)` otherwise.
     *
     * **Per-Operation Cancellation**
     *
     * All. Upon cancellation, the asynchronous operation continues to run.
     */
    template <class CompletionToken = agrpc::DefaultCompletionToken>
    auto wait(CompletionToken token = {})
    {
        assert(!completion_handler_ && "Can only wait again when the previous wait has been cancelled or completed");
        return asio::async_initiate<CompletionToken, CompletionSignature>(Initiator{*this}, token);
    }

  private:
    struct Initiator
    {
        template <class CompletionHandler>
        void operator()(CompletionHandler&& ch)
        {
            if (self_.result_)
            {
                auto executor = asio::get_associated_executor(ch);
                const auto allocator = asio::get_associated_allocator(ch);
                auto local_result{static_cast<Result&&>(*self_.result_)};
                self_.result_.reset();
                detail::post_with_allocator(
                    std::move(executor),
                    [ch = static_cast<CompletionHandler&&>(ch),
                     local_result = static_cast<Result&&>(local_result)]() mutable
                    {
                        detail::invoke_successfully_from_tuple(std::move(ch), static_cast<Result&&>(local_result));
                    },
                    allocator);
                return;
            }
            auto cancellation_slot = asio::get_associated_cancellation_slot(ch);
            self_.emplace_completion_handler(static_cast<CompletionHandler&&>(ch));
            self_.install_cancellation_handler(cancellation_slot);
        }

        CancelSafe& self_;
    };

    struct CancellationHandler
    {
        void operator()(asio::cancellation_type type)
        {
            if (static_cast<bool>(type & asio::cancellation_type::all))
            {
                if (auto ch = self_.completion_handler_.release())
                {
                    detail::complete_operation_aborted(std::move(ch), CompletionArgs{}...);
                }
            }
        }

        CancelSafe& self_;
    };

    template <class CompletionHandler>
    void emplace_completion_handler(CompletionHandler&& ch)
    {
        completion_handler_
            .template emplace<detail::WorkTrackingCompletionHandler<detail::RemoveCrefT<CompletionHandler>>>(
                static_cast<CompletionHandler&&>(ch));
    }

    template <class CancellationSlot>
    void install_cancellation_handler(CancellationSlot& cancellation_slot)
    {
        if (cancellation_slot.is_connected())
        {
            cancellation_slot.assign(CancellationHandler{*this});
        }
    }

    detail::AtomicTypeErasedCompletionHandler<CompletionSignature> completion_handler_{};
    std::optional<Result> result_;
};

/**
 * @brief CancelSafe templated on `void(bool)`
 */
using GrpcCancelSafe = agrpc::CancelSafe<void(bool)>;

AGRPC_NAMESPACE_END

#endif

#endif  // AGRPC_AGRPC_CANCEL_SAFE_HPP
